@@ -8,6 +8,8 @@
  * Copyright (C) 2000-2002, Ximian, Inc.
  */
 
+#include <string.h>
+
 #include "soup-auth.h"
 #include "soup-error.h"
 #include "soup-message.h"
@@ -147,6 +149,9 @@ release_connection (const SoupDataBuffer *data,
 {
 	SoupConnection *conn = user_data;
 	soup_connection_release (conn);
+
+	if (data->owner == SOUP_BUFFER_SYSTEM_OWNED)
+		g_free (data->body);
 }
 
 static void 
@@ -180,23 +185,19 @@ soup_message_cleanup (SoupMessage *req)
 						  release_connection,
 						  release_and_close_connection,
 						  req->connection);
-		req->priv->read_tag = 0;
+		soup_transfer_read_unref (req->priv->read_tag);
+		req->priv->read_tag = NULL;
 		req->connection = NULL;
-		/* 
-		 * The buffer doesn't belong to us until the message is 
-		 * finished.
-		 */
-		req->response.owner = SOUP_BUFFER_STATIC;
 	}
 
 	if (req->priv->read_tag) {
 		soup_transfer_read_cancel (req->priv->read_tag);
-		req->priv->read_tag = 0;
+		req->priv->read_tag = NULL;
 	}
 
 	if (req->priv->write_tag) {
 		soup_transfer_write_cancel (req->priv->write_tag);
-		req->priv->write_tag = 0;
+		req->priv->write_tag = NULL;
 	}
 
 	if (req->priv->connect_tag) {
@@ -522,39 +523,10 @@ soup_message_queue (SoupMessage    *req,
 	soup_queue_message (req, callback, user_data);
 }
 
-typedef struct {
-	SoupMessage *msg;
-	SoupAuth    *conn_auth;
-} RequeueConnectData;
-
-static void
-requeue_connect_cb (SoupContext          *ctx,
-		    SoupConnectErrorCode  err,
-		    SoupConnection       *conn,
-		    gpointer              user_data)
-{
-	RequeueConnectData *data = user_data;
-
-	if (conn && !conn->auth)
-		conn->auth = data->conn_auth;
-	else
-		soup_auth_free (data->conn_auth);
-
-	soup_queue_connect_cb (ctx, err, conn, data->msg);
-	if (data->msg->errorcode)
-		soup_message_issue_callback (data->msg);
-
-	g_free (data);
-}
-
 static void
 requeue_read_error (gboolean body_started, gpointer user_data)
 {
-	RequeueConnectData *data = user_data;
-	SoupMessage *msg = data->msg;
-	SoupContext *dest_ctx = msg->connection->context;
-
-	soup_context_ref (dest_ctx);
+	SoupMessage *msg = user_data;
 
 	soup_connection_set_keep_alive (msg->connection, FALSE);
 	soup_connection_release (msg->connection);
@@ -563,29 +535,22 @@ requeue_read_error (gboolean body_started, gpointer user_data)
 	soup_queue_message (msg, 
 			    msg->priv->callback, 
 			    msg->priv->user_data);
-
-	msg->status = SOUP_STATUS_CONNECTING;
-
-	msg->priv->connect_tag =
-		soup_context_get_connection (dest_ctx, 
-					     requeue_connect_cb, 
-					     data);
-
-	soup_context_unref (dest_ctx);
 }
 
 static void
 requeue_read_finished (const SoupDataBuffer *buf,
 		       gpointer        user_data)
 {
-	RequeueConnectData *data = user_data;
-	SoupMessage *msg = data->msg;
+	SoupMessage *msg = user_data;
 	SoupConnection *conn = msg->connection;
 
+	if (buf->owner == SOUP_BUFFER_SYSTEM_OWNED)
+		g_free (buf->body);
+
+	soup_connection_set_used (msg->connection);
 	if (!soup_connection_is_keep_alive (msg->connection))
-		requeue_read_error (FALSE, data);
+		requeue_read_error (FALSE, msg);
 	else {
-		g_free (data);
 		msg->connection = NULL;
 
 		soup_queue_message (msg, 
@@ -608,19 +573,19 @@ soup_message_requeue (SoupMessage *req)
 	g_return_if_fail (req != NULL);
 
 	if (req->connection && req->connection->auth && req->priv->read_tag) {
-		RequeueConnectData *data = NULL;
-
-		data = g_new0 (RequeueConnectData, 1);
-		data->msg = req;
-		data->conn_auth = req->connection->auth;
-
 		soup_transfer_read_set_callbacks (req->priv->read_tag,
 						  NULL,
 						  NULL,
 						  requeue_read_finished,
 						  requeue_read_error,
-						  data);
-		req->priv->read_tag = 0;
+						  req);
+		soup_transfer_read_unref (req->priv->read_tag);
+		req->priv->read_tag = NULL;
+
+		if (req->priv->write_tag) {
+			soup_transfer_write_cancel (req->priv->write_tag);
+			req->priv->write_tag = NULL;
+		}
 	} else
 		soup_queue_message (req, 
 				    req->priv->callback, 
@@ -660,202 +625,62 @@ soup_message_send (SoupMessage *msg)
 }
 
 static void
-maybe_validate_auth (SoupMessage *msg, gpointer user_data)
-{
-	gboolean proxy = GPOINTER_TO_INT (user_data);
-	int auth_failure;
-	SoupContext *ctx;
-	SoupAuth *auth;
-
-	if (proxy) {
-		ctx = soup_get_proxy ();
-		auth_failure = SOUP_ERROR_PROXY_UNAUTHORIZED; /* 407 */
-	}
-	else {
-		ctx = msg->context;
-		auth_failure = SOUP_ERROR_UNAUTHORIZED; /* 401 */
-	}
-
-	auth = soup_auth_lookup (ctx);
-	if (!auth)
-		return;
-
-	if (msg->errorcode == auth_failure) {
-		/* Pass through */
-	}
-	else if (msg->errorclass == SOUP_ERROR_CLASS_SERVER_ERROR) {
-		/* 
-		 * We have no way of knowing whether our auth is any good
-		 * anymore, so just invalidate it and start from the
-		 * beginning next time.
-		 */
-		soup_auth_invalidate (auth, ctx);
-	}
-	else {
-		auth->status = SOUP_AUTH_STATUS_SUCCESSFUL;
-		auth->controlling_msg = NULL;
-	}
-} /* maybe_validate_auth */
-
-static void 
 authorize_handler (SoupMessage *msg, gboolean proxy)
 {
-	const GSList *vals;
-	SoupAuth *auth;
 	SoupContext *ctx;
-	const SoupUri *uri;
-
-	if (msg->connection->auth &&
-	    msg->connection->auth->status == SOUP_AUTH_STATUS_SUCCESSFUL)
-		goto THROW_CANT_AUTHENTICATE;
 
 	ctx = proxy ? soup_get_proxy () : msg->context;
-	uri = soup_context_get_uri (ctx);
-
-	vals = soup_message_get_header_list (msg->response_headers, 
-					     proxy ? 
-					             "Proxy-Authenticate" : 
-					             "WWW-Authenticate");
-	if (!vals) goto THROW_CANT_AUTHENTICATE;
-
-	auth = soup_auth_lookup (ctx);
-	if (auth) {
-		g_assert (auth->status != SOUP_AUTH_STATUS_INVALID);
-
-		if (auth->status == SOUP_AUTH_STATUS_PENDING) {
-			if (auth->controlling_msg == msg) {
-				auth->status = SOUP_AUTH_STATUS_FAILED;
-				goto THROW_CANT_AUTHENTICATE;
-			}
-			else {
-				soup_message_requeue (msg);
-				return;
-			}
-		}
-		else if (auth->status == SOUP_AUTH_STATUS_FAILED ||
-			 auth->status == SOUP_AUTH_STATUS_SUCCESSFUL) {
-			/*
-			 * We've failed previously, but we'll give it
-			 * another go, or we've been successful
-			 * previously, but it's not working anymore.
-			 *
-			 * Invalidate the auth, so it's removed from the
-			 * hash and try it again as if we never had it
-			 * in the first place.
-			 */
-			soup_auth_invalidate (auth, ctx);
-			soup_message_requeue (msg);
-			return;
-		}
-	}
-
-	if (!auth) {
-		auth = soup_auth_new_from_header_list (uri, vals);
-
-		if (!auth) {
-			soup_message_set_error_full (
-				msg, 
-				proxy ? 
-			                SOUP_ERROR_CANT_AUTHENTICATE_PROXY : 
-			                SOUP_ERROR_CANT_AUTHENTICATE,
-				proxy ? 
-			                "Unknown authentication scheme "
-				        "required by proxy" :
-			                "Unknown authentication scheme "
-				        "required");
-			return;
-		}
-
-		auth->status = SOUP_AUTH_STATUS_PENDING;
-		auth->controlling_msg = msg;
-		soup_message_add_handler (msg, SOUP_HANDLER_PRE_BODY,
-					  maybe_validate_auth,
-					  GINT_TO_POINTER (proxy));
-	}
-
-	/*
-	 * Call registered authenticate handler
-	 */
-	if (!uri->user && soup_auth_fn)
-		(*soup_auth_fn) (auth->type,
-				 (SoupUri *) uri,
-				 auth->realm, 
-				 soup_auth_fn_user_data);
-
-	if (!uri->user) {
-		soup_auth_free (auth);
-		goto THROW_CANT_AUTHENTICATE;
-	}
-
-	/*
-	 * Initialize with auth data (possibly returned from 
-	 * auth callback).
-	 */
-	soup_auth_initialize (auth, uri);
-
-	if (auth->type == SOUP_AUTH_TYPE_NTLM) {
-		SoupAuth *old_auth = msg->connection->auth;
-
-		if (old_auth)
-			soup_auth_free (old_auth);
-		msg->connection->auth = auth;
-	} else
-		soup_auth_set_context (auth, ctx);
-
-	soup_message_requeue (msg);
-
-        return;
-
- THROW_CANT_AUTHENTICATE:
-	soup_message_set_error (msg, 
-				proxy ? 
+	if (soup_context_update_auth (ctx, msg))
+		soup_message_requeue (msg);
+	else {
+		soup_message_set_error (msg, 
+					proxy ? 
 			                SOUP_ERROR_CANT_AUTHENTICATE_PROXY : 
 			                SOUP_ERROR_CANT_AUTHENTICATE);
+	}
 }
 
 static void 
 redirect_handler (SoupMessage *msg, gpointer user_data)
 {
 	const gchar *new_loc;
+	const SoupUri *old_uri;
+	SoupUri *new_uri;
+	SoupContext *new_ctx;
 
-	if (msg->errorclass != SOUP_ERROR_CLASS_REDIRECT || 
-	    msg->priv->msg_flags & SOUP_MESSAGE_NO_REDIRECT) return;
+	if (msg->priv->msg_flags & SOUP_MESSAGE_NO_REDIRECT)
+		return;
 
 	new_loc = soup_message_get_header (msg->response_headers, "Location");
+	if (!new_loc)
+		return;
 
-	if (new_loc) {
-		const SoupUri *old_uri;
-		SoupUri *new_uri;
-		SoupContext *new_ctx;
+	old_uri = soup_context_get_uri (msg->context);
 
-		old_uri = soup_context_get_uri (msg->context);
+	new_uri = soup_uri_new (new_loc);
+	if (!new_uri) 
+		goto INVALID_REDIRECT;
 
-		new_uri = soup_uri_new (new_loc);
-		if (!new_uri) 
-			goto INVALID_REDIRECT;
+	/* 
+	 * Copy auth info from original URI.
+	 */
+	if (old_uri->user && !new_uri->user)
+		soup_uri_set_auth (new_uri,
+				   old_uri->user, 
+				   old_uri->passwd, 
+				   old_uri->authmech);
 
-		/* 
-		 * Copy auth info from original URI.
-		 */
-		if (old_uri->user && !new_uri->user)
-			soup_uri_set_auth (new_uri,
-					   old_uri->user, 
-					   old_uri->passwd, 
-					   old_uri->authmech);
+	new_ctx = soup_context_from_uri (new_uri);
 
-		new_ctx = soup_context_from_uri (new_uri);
+	soup_uri_free (new_uri);
 
-		soup_uri_free (new_uri);
+	if (!new_ctx)
+		goto INVALID_REDIRECT;
 
-		if (!new_ctx)
-			goto INVALID_REDIRECT;
-
-		soup_message_set_context (msg, new_ctx);
-		soup_context_unref (new_ctx);
-
-		soup_message_requeue (msg);
-	}
+	soup_message_set_context (msg, new_ctx);
+	soup_context_unref (new_ctx);
 
+	soup_message_requeue (msg);
 	return;
 
  INVALID_REDIRECT:
@@ -891,8 +716,8 @@ static SoupHandlerData global_handlers [] = {
 		SOUP_HANDLER_PRE_BODY,
 		redirect_handler, 
 		NULL, 
-		RESPONSE_HEADER_HANDLER, 
-		{ (guint) "Location" }
+		RESPONSE_ERROR_CLASS_HANDLER, 
+		{ SOUP_ERROR_CLASS_REDIRECT }
 	},
 	/* 
 	 * Handle authorization.
